require 'spec_helper'
require 'live_management'
require 'resource_management'
require 'rack/test'
require 'json'
require 'digest/md5'
LiveManagement.disable :show_exceptions

describe 'the LiveManagement REST API' do
  include Rack::Test::Methods

  def app
    LiveManagement
  end

  before do
    @emcee = Emcee.new
    Emcee.stubs(:new).returns(@emcee)

    @nodes = ['node1.foo.bar', 'node2.bar.baz', 'node3', 'whatever' ].map {|name| stub(:name => name) }
    @emcee.stubs(:nodes).returns(@nodes)
  end

  describe 'GET /lm/resource_types' do
    it 'should return a list of JSON hashes with names of resource types' do
      get "/lm/resource_types"
      result = JSON.parse(last_response.body)
      result.each do |r|
        r.keys.first.should == "name"
        ['user','package','host','group','service'].should include(r["name"])
      end
    end
  end

  describe 'GET /lm/resources/:type' do
    let(:resource_type) {'goat'}

    describe 'and a nodeset id is passed' do
      it 'should filter results to that nodeset' do
        nodes = ['foo', 'bar']
        Emcee.expects(:new).with(all_of(has_entry(:nodeset => nodes), Not(has_key(:filter)))).returns(@emcee)
        @emcee.stubs(:resources).returns([])
        ResourceManagement.expects(:retrieve_nodeset).returns(nodes)

        get "/lm/resources/#{resource_type}", { :nodeset_id => '1234' }

        result = JSON.parse(last_response.body)
        result['resources'].should == []
        result['total'].should == 0
        result['variations'].should == 0
        last_response.should be_ok
      end

      it 'and the nodeset is empty should return an empty resource set' do
        Emcee.expects(:new).never
        ResourceManagement.expects(:retrieve_nodeset).returns([])

        get "/lm/resources/#{resource_type}", { :nodeset_id => '1234' }

        result = JSON.parse(last_response.body)
        result['resources'].should == []
        result['total'].should == 0
        result['variations'].should == 0
        last_response.should be_ok
      end

      it 'and no corresponding nodeset exists should respond with a 404 error' do
        get "/lm/resources/#{resource_type}", { :nodeset_id => '1234' }

        last_response.body.should =~ /No nodeset was found/
        last_response.should_not be_ok
      end
    end

    describe 'but there are no resources of the specified type' do
      before do
        @emcee.stubs(:resources).returns([])
      end

      it 'should return a JSON hash with an empty list of resources' do
        ResourceManagement.expects(:retrieve_nodeset).returns([])
        get "/lm/resources/#{resource_type}", { :nodeset_id => '1234' }

        last_response.should be_ok
        result = JSON.parse(last_response.body)
        result['resources'].should == []
        result['total'].should == 0
        result['variations'].should == 0
      end
    end

    describe 'and there are matching resources of the specified type' do
      let(:goats){[
          { 'title' => 'jerry',
            'parameters' => {'favoritefood' => 'cans', 'furcolor' => 'zebra'},
            'type' => resource_type, 'node_name' => 'node1' },
          { 'title' => 'jerry',
            'parameters' => {'favoritefood' => 'rocks', 'furcolor' => 'zebra'},
            'type' => resource_type, 'node_name' => 'node2' },
          { 'title' => 'bingo',
            'parameters' => {'favoritefood' => 'dirt', 'furcolor' => 'calico'},
            'type' => resource_type, 'node_name' => 'node2' },
          { 'title' => 'reginald',
            'parameters' => {'favoritefood' => 'brie', 'furcolor' => 'tortoiseshell'},
            'type' => resource_type, 'node_name' => 'node3' },
          { 'title' => 'reginald',
            'parameters' => {'favoritefood' => 'brie', 'furcolor' => 'tortoiseshell'},
            'type' => resource_type, 'node_name' => 'node1' }
        ]}

      before do
        @emcee.stubs(:resources).returns(goats)
      end

      it 'should return a JSON list of hashes with resource variation information' do
        ResourceManagement.expects(:retrieve_nodeset).returns(['node1', 'node2', 'node3'])
        get "/lm/resources/#{resource_type}", { :nodeset_id => '1234' }
        results = JSON.parse(last_response.body)
        results['total'].should            == 5
        results['variations'].should       == 4
        results['resources'].length.should == 3

        resources = results['resources'].sort_by { |r| r['title'] }
        resources[0].to_a.should =~ {
          'title'      => 'bingo',
          'nodeCount'  => 1,
          'properties' => ['furcolor', 'favoritefood'],
          'variations' => [{ 'attributes' => {'favoritefood' => 'dirt', 'furcolor' => 'calico'},
                             'nodes' => ['node2']}]
        }.to_a

        # extract variations to assert on its own to avoid sort issues
        r1_variations = resources[1].delete('variations')
        resources[1].to_a.should =~ {
          'title'      => 'jerry',
          'nodeCount'  => 2,
          'properties' => ['furcolor', 'favoritefood']
        }.to_a
        r1_variations.should =~ [
          {'attributes' => {'favoritefood' => 'cans', 'furcolor' => 'zebra'}, 'nodes' =>['node1']},
          {'attributes' => {'favoritefood' => 'rocks', 'furcolor' => 'zebra'}, 'nodes' => ['node2']}
        ]

        resources[2].to_a.should =~ {
          'title'      => 'reginald',
          'nodeCount'  => 2,
          'properties' => ['furcolor', 'favoritefood'],
          'variations'=> [{'attributes' => {'favoritefood' => 'brie', 'furcolor' => 'tortoiseshell'},
                           'nodes' => ['node3', 'node1']}]
        }.to_a
      end
    end
  end

  describe 'GET /lm/resources/:type/:title' do
    let(:resource_type){'goat'}
    let(:title){'jerry'}

    describe 'and a nodeset id is passed' do
      it 'and the nodeset is empty should return an empty resource set' do
        Emcee.expects(:new).never
        ResourceManagement.expects(:retrieve_nodeset).returns([])

        get "/lm/resources/#{resource_type}/#{title}", { :nodeset_id => '1234' }

        result = JSON.parse(last_response.body)
        result.should == {"title" => "jerry",
                          "variations" => [],
                          "nodeCount" => 0,
                          "properties" => []}
        last_response.should be_ok
      end
    end

    describe 'when no nodeset is specified' do
      describe 'and there are no matching resources' do
        before do
          @emcee.stubs(:resource).returns([])
        end

        it 'should error with a 400' do
          get "/lm/resources/#{resource_type}/#{title}"
          results = last_response.body
          last_response.status.should == 400
          last_response.body.should == 'Missing required parameter: [ nodeset_id ]'
        end
      end

      describe 'and there are matching resources' do
        let(:goats) do [
          { 'title' => 'jerry',
            'parameters' => {'waterholder' => 'cans', 'furcolor' => 'zebra'},
            'type' => resource_type, 'node_name' => 'node1' },
          { 'title' => 'jerry',
            'parameters' => {'favoritefood' => 'rocks', 'furcolor' => 'zebra'},
            'type' => resource_type, 'node_name' => 'node2' },
          { 'title' => 'jerry',
            'parameters' => {'favoritefood' => 'rocks', 'furcolor' => 'zebra'},
            'type' => resource_type, 'node_name' => 'node3' },
          ]
        end

        before do
          @emcee.stubs(:resource).returns(goats)
        end

        describe 'and a nodeset id is passed' do
          it 'should filter results to that nodeset' do
            nodes = ['foo', 'bar']
            Emcee.expects(:new).with(all_of(has_entry(:nodeset => nodes), Not(has_key(:filter)))).returns(@emcee)
            ResourceManagement.expects(:retrieve_nodeset).returns(nodes)

            get "/lm/resources/#{resource_type}/#{title}", { :nodeset_id => '1234' }

            results = last_response.body
            last_response.should be_ok
          end
        end
        it "should return a JSON list with the node's resources and their data" do
          ResourceManagement.expects(:retrieve_nodeset).returns(['node1', 'node2', 'node3'])
          get "/lm/resources/#{resource_type}/#{title}", { :nodeset_id => '1234' }

          results = JSON.parse(last_response.body)

          variations = results.delete('variations')
          results.to_a.should =~ {
            'title'      => 'jerry',
            'nodeCount'  => 3,
            'properties' => ['furcolor', 'favoritefood', 'waterholder'],
          }.to_a
          variations.should =~ [
            {'attributes' => {'favoritefood' => 'rocks', 'furcolor' => 'zebra'},
             'nodes' => ['node2', 'node3']},
            {'attributes' => {'waterholder' => 'cans', 'furcolor' => 'zebra'},
             'nodes' => ['node1']}
          ]

          last_response.should be_ok
        end
      end
    end
  end

  describe 'POST /lm/mc_run' do

    before do
      @nodeset_id = '1598ba21f'
      @agent_name = 'foo'
      @action_name = 'purge'
      @input = {'input1' => 'tupni'}
      @agent = Emcee::Agent.new(nil, {:agent => @agent_name, :version => 0.1})
      @action = Emcee::Action.new(@agent, fake_interfaces[@action_name])
    end

    describe 'called with nodeset_id, agent_name, action_name, and input' do
      it 'calls the appropriate action' do
        @action.expects(:run).with({:input1 => 'tupni'}).returns([])
        @agent.expects(:action_by_name).with(@action_name).returns(@action)
        @emcee.expects(:agents_by_name).with(@agent_name).returns([@agent])
        ResourceManagement.expects(:retrieve_nodeset).with(@nodeset_id).returns(['node1'])
        post "lm/mc_run", {
          :nodeset_id  => @nodeset_id,
          :agent_name  => @agent_name,
          :action_name => @action_name,
          :input       => @input
        }
        result = JSON.parse(last_response.body)
        result.should == []
      end
    end

    describe 'called with nodeset_id, agent_name, and action_name, but not input' do
      it 'calls the appropriate action' do
        @action.expects(:run).with().returns([])
        @agent.expects(:action_by_name).with(@action_name).returns(@action)
        @emcee.expects(:agents_by_name).with(@agent_name).returns([@agent])
        ResourceManagement.expects(:retrieve_nodeset).with(@nodeset_id).returns(['node1'])
        post "lm/mc_run", {
          :nodeset_id  => @nodeset_id,
          :agent_name  => @agent_name,
          :action_name => @action_name,
        }
        result = JSON.parse(last_response.body)
        result.should == []
      end
    end

  end

  describe 'POST /lm/clone' do
    before do
      @resource = { :title => 'jerry', :parameters => {"favoritefood" => 'cans', "furcolor" => 'zebra'}, :type => 'goat', :node_name => 'node1' }
      @successful_result = {}
      @resource.each { |k,v| @successful_result[k.to_s] = v }
      @successful_result['node_name'] = 'node2'
      @origin_nodes = ['node1', 'node5', 'node6']
      @target_nodes = ['node1', 'node2', 'node3', 'node4']
      ResourceManagement.expects(:retrieve_nodeset).at_most_once.returns(@target_nodes)
      results = [stub('result0'), stub('result1'), stub('result2')]
      results[0].stubs(:node_name).returns("node2")
      results[0].stubs(:status_code).returns(0)
      results[0].stubs(:raw_contents).returns({ :status   => "Resource was created", 
                                                :resource => @successful_result})
      results[0].stubs(:status_message).returns("OK")
      results[1].stubs(:node_name).returns("node3")
      results[1].stubs(:status_code).returns(1)
      results[1].stubs(:status_message).returns("Error!")
      results[2].stubs(:node_name).returns("node4")
      results[2].stubs(:status_code).returns(0)
      results[2].stubs(:raw_contents).returns({ :status   => "Resource was not created." })
      results[2].stubs(:contents).returns("Resource was not created.")
      results[2].stubs(:status_message).returns("OK")
      @emcee.stubs(:create_resource).returns(results)
    end

    describe 'called with type, parameters, target nodes, and origin nodes' do
      describe 'but not with title' do
        it 'should fail with a 400 response code and an error message' do
          post "lm/clone", {
            :type         => @resource[:type],
            :parameters   => @resource[:parameters],
            :nodeset_id   => '1234',
            :origin_nodes => @origin_nodes
          }
          last_response.should_not be_ok
          last_response.status.should == 400
          last_response.body.should == 'Missing required parameter: [ title ]'
        end
      end

      describe 'and also with title' do
        it 'should not attempt to clone to the origin nodes' do
          ResourceManagement.
            expects(:clone).
            with(@target_nodes - @origin_nodes,
                 { 'type'       => @resource[:type],
                   'title'      => @resource[:title],
                   'parameters' => @resource[:parameters] }).
            returns( {:successes => [], :failures => [], :errors => []} )
          post "lm/clone", {
            :title      => @resource[:title],
            :type       => @resource[:type],
            :parameters => @resource[:parameters],
            :nodeset_id => '1234',
            :origin_nodes => @origin_nodes
          }
        end
        it 'should attempt to create a copy of the resource' do
          post "lm/clone", {
            :title      => @resource[:title],
            :type       => @resource[:type],
            :parameters => @resource[:parameters],
            :nodeset_id => '1234',
            :origin_nodes => @origin_nodes
          }
          result = JSON.parse(last_response.body)
          result["failures"].should ==
            [{ "nodeName"    => "node4",
            "reason"         => "Resource was not created." }]
          result["successes"].should ==
            [ { 'attributes' => @successful_result['parameters'],
                'nodes'      => [@successful_result['node_name']]
              } ]
          result["errors"].should ==
            [{ "nodeName"    => "node3",
            "reason" => "Error!"}]
          result["counts"].should == { 'successes' => 1, 'failures' => 1, 'errors' => 1 }
          result.has_key?("original").should be_true
          result["properties"].should == ["favoritefood", "furcolor"]
        end
      end

      describe 'and clone results include resources of a different type or title' do

        it 'should raise an error' do
          ResourceManagement.expects(:clone).returns({ :successes => ['foo', 'bar'],
                                                       :failures  => [],
                                                       :errors    => [] })
          post "lm/clone", {
            :title      => @resource[:title],
            :type       => @resource[:type],
            :parameters => @resource[:parameters],
            :nodeset_id => '1234',
            :origin_nodes => @origin_nodes
          }
          last_response.body.should == "Something wildly unexpected happened. Please file a bug at http://projects.puppetlabs.com/projects/puppet-enterprise/issues"
          last_response.status.should == 500
        end

      end
    end
  end

  describe 'POST /lm/clone/preview' do
    let(:resource_type){'goat'}
    let(:title){'jerry'}
    let(:goats) do [
      stub({ :contents => {'title' => 'jerry', :managed => true,
        'parameters' => {'waterholder' => 'cans', 'furcolor' => 'zebra'},
        'type' => resource_type}, :node_name => 'node1' }),
      stub({ :contents => {'title' => 'jerry', :managed => false,
        'parameters' => {'favoritefood' => 'rocks', 'furcolor' => 'zebra'},
        'type' => resource_type}, :node_name => 'node2' }),
      stub({ :contents => {'title' => 'jerry', :managed => 'unknown',
        'parameters' => {'leastfavoritefood' => 'rocks', 'furcolor' => 'zebra'},
        'type' => resource_type}, :node_name => 'node3' }),
      stub({ :contents => {'title' => 'jerry', :managed => true,
        'parameters' => {:ensure => :absent, 'furcolor' => 'zebra'},
        'type' => resource_type}, :node_name => 'node4' }),
      ]
    end
    let(:origin_nodes){['node2', 'node6']}

    before do
      # XXX : This test is overly stubbed.
      Emcee::Resource.
        stubs(:run_ral_action).
        with(@emcee, 'find', {:type => resource_type, :title => title}).
        returns goats.reject { |v| v.node_name == "node2" }
    end

    it 'should return an empty set of results if the origin and target node sets are the same' do
      ResourceManagement.expects(:retrieve_nodeset).returns(origin_nodes)
      post "/lm/clone/preview", {
        :nodeset_id   => '1234',
        :type         => resource_type,
        :title        => title,
        :parameters   => {'special' => 'yes'},
        :origin_nodes => origin_nodes,
      }
      result = JSON.parse(last_response.body)
      result.should == {
        'creations' => [],
        'failures' => [],
        'overwrites' => [],
        'properties' => [],
        'counts' => {
          'creations' => 0,
          'failures' => 0,
          'overwrites' => 0
        }
      }
    end

    it 'should return a list of potential successes, overwrites and failures' do
      ResourceManagement.expects(:retrieve_nodeset).returns(['node1', 'node2', 'node3', 'node4', 'node5'])
      post "/lm/clone/preview", {
        :nodeset_id   => '1234',
        :type         => resource_type,
        :title        => title,
        :parameters   => {'special' => 'yes'},
        :origin_nodes => origin_nodes,
      }
      result = JSON.parse(last_response.body)
      result['creations'].should == [ { 'nodeName' => 'node5' } ]
      result['failures'].should == [
        { 'nodeName' => 'node1',
          'reason' => "Resource is managed by Puppet." },
        { 'nodeName' => 'node4',
          'reason' => "Resource is managed by Puppet." }
      ]
      result['overwrites'].should =~ [
          {"attributes"=>{"furcolor"=>"zebra", "leastfavoritefood"=>"rocks"},
           "nodes"=>["node3"]}
      ]
      result['properties'].should =~ [
      "leastfavoritefood", "furcolor", "special"
      ]
      result['counts'].should == {
        'creations'  => 1,
        'failures'   => 2,
        'overwrites' => 1,
      }
    end
  end

  describe 'GET /lm/nodes' do
    let(:nodes) { ["testhost2", "testhost1", "testhost3"] }
    let(:nodeset_id) { Digest::MD5.new.hexdigest(nodes.sort.join) }

    describe 'and no filter is passed' do
      it 'should return all known hostnames with nodeset_id' do
        @emcee.expects(:filtered_nodes).returns(nodes)

        get 'lm/nodes'

        last_response.body.should == {
          'id'    => nodeset_id,
          'nodes' => nodes.sort.map {|n| {'name' => n} }
        }.to_json
        last_response.should be_ok
      end
    end

    describe 'and a filter is passed' do
      it 'should return a list of hostnames matching the submitted filters' do
        @emcee.expects(:filtered_nodes).returns(nodes)
        params = {
          :facts      => {'name' => 'foo', 'value' => 'bar'},
          :classes    => "bazclass",
          :identities => "bazinga.localhost.baz"
        }

        get 'lm/nodes', params

        last_response.body.should == {
          'id'    => nodeset_id,
          'nodes' => nodes.sort.map {|n| {'name' => n} }
        }.to_json
        last_response.should be_ok
      end

      it 'should fail with a 400 response if the params are malformed' do
        params = {:banana => 'delicious'}
        get 'lm/nodes', params
        last_response.body.should == 'Unknown parameter: [ banana ]'
        last_response.status.should == 400
      end

      it 'should be able to handle empty params' do
        @emcee.expects(:filtered_nodes).returns(nodes)
        params = {
          :facts      => {'name' => '', 'value' => ''},
          :classes    => '',
          :identities => ''
        }

        get 'lm/nodes', params

        last_response.body.should == {
          'id'    => nodeset_id,
          'nodes' => nodes.sort.map {|n| {'name' => n} }
        }.to_json
        last_response.should be_ok
      end

    end
  end

  describe '/lm/nodeset' do
    before do
      @nodes = ["node1", "node3", "node2"]
    end

    after do
      Emcee::Cache.new(Digest::MD5.new.hexdigest(@nodes.sort.join)).expire!
    end

    describe 'when POSTing' do
      describe 'and no set of nodes is passed' do
        it 'should return a nodeset ID for the empty list' do
          nodesetid = Digest::MD5.new.hexdigest([].join)
          post 'lm/nodeset'
          last_response.body.should == nodesetid
          last_response.should be_ok
        end
      end
      describe 'and a set of nodes is passed' do
        describe 'and the requested set of nodes is cached' do
          it 'should retrieve and return the nodeset id' do
            nodesetid = Digest::MD5.new.hexdigest(@nodes.sort.join)
            Emcee::Cache.new(nodesetid).save(@nodes)
            post 'lm/nodeset', :nodes => @nodes
            last_response.should be_ok
            last_response.body.should == nodesetid

            get 'lm/nodeset', :key => last_response.body
            last_response.body.should == ["node1", "node2", "node3"].to_json
          end
        end

        describe 'and the requested set of nodes is not cached' do
          it 'should generate a nodeset ID and store the nodes' do
            post 'lm/nodeset', :nodes => @nodes
            last_response.should be_ok
            last_response.body.should == Digest::MD5.new.hexdigest(@nodes.sort.join)

            get 'lm/nodeset', :key => last_response.body
            last_response.body.should == ["node1", "node2", "node3"].to_json
          end
        end
      end
    end

    describe 'when GETing' do
      describe 'and the requested nodeset is cached' do
        it 'should return the list of nodes in the set' do
          nodeset_id = Digest::MD5.new.hexdigest(@nodes.sort.join)
          Emcee::Cache.new(nodeset_id).save(@nodes)
          get 'lm/nodeset', :key => nodeset_id
          last_response.body.should == @nodes.sort.to_json
          last_response.should be_ok
        end
      end

      describe 'and the requested nodeset is not cached' do
        it 'should fail with a 400 error code' do
          get 'lm/nodeset', :key => Digest::MD5.new.hexdigest(@nodes.sort.join)
          last_response.body.should =~ /No such nodeset is cached/
          last_response.status.should == 400
        end
      end
    end
  end

end
